#include "../include/mptp.h"
#include "../include/error.h"
#include "../include/url.h"
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/socket.h>
#include <unistd.h>

bool lm_mptp_init(lm_mptp_t *packet, bool is_request, uint8_t code, bool is_last) {
  packet->data  = NULL;
  packet->flags = 0;
  packet->size  = 0;

  if (code > MPTP_CODE_MAX) {
    lm_error_set(LM_ERR_MPTPBadCode);
    return false;
  }

  // X-X-X-X-0-0-0-0
  packet->flags |= (MPTP_VERSION_SUPPORTED << 4);

  // 0-0-0-0-X-0-0-0
  if (is_request)
    packet->flags |= (MPTP_REQUEST << 3);
  else
    packet->flags |= (MPTP_RESPONSE << 3);

  // 0-0-0-0-0-X-X-0
  packet->flags |= (code << 1);

  // 0-0-0-0-0-0-0-X
  if (is_last)
    packet->flags |= 1;
  else
    packet->flags |= 0;

  return true;
}

int lm_mptp_connect(lm_url_t *url) {
  if (NULL == url || NULL == url->host) {
    lm_error_set(LM_ERR_MPTPBadUrl);
    return -1;
  }

  int                sock;
  struct sockaddr_in addr;
  struct hostent    *ent = gethostbyname(url->host);
  if (NULL == ent) {
    lm_error_set(LM_ERR_MPTPHostFail);
    return -1;
  }

  if (ent->h_addrtype != AF_INET && ent->h_addrtype != AF_INET6) {
    lm_error_set(LM_ERR_MPTPHostFail);
    return -1;
  }

  if (NULL == ent->h_addr_list[0]) {
    lm_error_set(LM_ERR_MPTPHostFail);
    return -1;
  }

  addr.sin_addr.s_addr = *(long *)(ent->h_addr_list[0]);
  addr.sin_family      = ent->h_addrtype;
  addr.sin_port        = htons(url->port);

  if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    lm_error_set(LM_ERR_MPTPSocketFail);
    return -1;
  }

  struct timeval timeout;
  bzero(&timeout, sizeof(timeout));

  timeout.tv_sec  = MPTP_TIMEOUT;
  timeout.tv_usec = 0;

  if (setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char *)&timeout, sizeof(timeout)) < 0) {
    lm_mptp_disconnect(sock);
    lm_error_set(LM_ERR_MPTPSetsockopt);
    return -1;
  }

  if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    lm_mptp_disconnect(sock);
    lm_error_set(LM_ERR_MPTPConnectFail);
    return -1;
  }

  return sock;
}

bool lm_mptp_send(int sock, lm_mptp_t *packet) {
  if (MPTP_FLAGS_VERSION(packet) != MPTP_VERSION_SUPPORTED) {
    lm_error_set(LM_ERR_MPTPBadVersion);
    return false;
  }

  size_t size = sizeof(packet->flags) + sizeof(packet->size);

  if (NULL == packet->data || packet->size <= 0) {
    if (send(sock, packet, size, 0) < 0) {
      lm_error_set(LM_ERR_MPTPSendFail);
      return false;
    }
    return true;
  }

  if (send(sock, packet, size, 0) < 0) {
    lm_error_set(LM_ERR_MPTPSendFail);
    return false;
  }

  if (send(sock, packet->data, packet->size, 0) < 0) {
    lm_error_set(LM_ERR_MPTPSendFail);
    return false;
  }

  return true;
}

bool lm_mptp_send_data(int sock, char *data, size_t size) {
  if (size > MPTP_CHUNK_MAX) {
    lm_error_set(LM_ERR_MPTPChunkFail);
    return false;
  }

  if (send(sock, data, size, 0) < 0) {
    lm_error_set(LM_ERR_MPTPSendFail);
    return false;
  }

  return true;
}

bool lm_mptp_recv(int sock, lm_mptp_t *packet) {
  if (recv(sock, &packet->flags, sizeof(packet->flags), 0) < 0)
    goto recvfail;

  if (recv(sock, &packet->size, sizeof(packet->size), 0) < 0)
    goto recvfail;

  if (MPTP_FLAGS_VERSION(packet) != MPTP_VERSION_SUPPORTED) {
    lm_error_set(LM_ERR_MPTPBadVersion);
    return false;
  }

  if (NULL == packet->data)
    return true;

  if (packet->size <= 0)
    return true;

  if (packet->size > MPTP_CHUNK_MAX) {
    lm_error_set(LM_ERR_MPTPChunkFail);
    return false;
  }

  if (recv(sock, &packet->data, packet->size, 0) < 0)
    goto recvfail;

  return true;

recvfail:
  if (errno == ETIMEDOUT || errno == EAGAIN)
    lm_error_set(LM_ERR_MPTPTimeout);
  else
    lm_error_set(LM_ERR_MPTPRecvFail);
  return false;
}

bool lm_mptp_recv_data(int sock, char *data, size_t size) {
  if (size > MPTP_CHUNK_MAX) {
    lm_error_set(LM_ERR_MPTPChunkFail);
    return false;
  }

  if (recv(sock, data, size, 0) < 0) {
    if (errno == ETIMEDOUT)
      lm_error_set(LM_ERR_MPTPTimeout);
    else
      lm_error_set(LM_ERR_MPTPRecvFail);
    return false;
  }

  return true;
}

void lm_mptp_disconnect(int sock) {
  close(sock);
}
